import logging

import altair as alt
import numpy as np
import pandas as pd
from pandas.api.types import (
    is_bool_dtype,
    is_categorical_dtype,
    is_datetime64_any_dtype,
    is_numeric_dtype,
    is_object_dtype,
)

from profiling.utils import vega_sanitize


def countplot(x_var: pd.Series):
    """Plot loss metric by quantile/category"""
    if is_categorical_dtype(x_var) or is_bool_dtype(x_var) or is_object_dtype(x_var):
        return countplot_cat(x_var)
    elif is_numeric_dtype(x_var):
        return countplot_num(x_var)
    elif is_datetime64_any_dtype(x_var):
        return countplot_datetime(x_var)
    else:
        logging.warning(
            f"Data type ({x_var.dtype}) not detected for {x_var.name}, will not generate count plot"
        )


def trim_outliers(x_var: pd.Series, n_std=4) -> pd.Series:
    """Remove values +/- n_std standard deviations from the mean."""
    mean = np.mean(x_var)
    std = np.std(x_var)

    return x_var[(x_var <= (mean + n_std * std)) & (x_var >= (mean - n_std * std))]


def get_bins(x_var: pd.Series, n_bins=25) -> pd.DataFrame:
    """Divide the numerical series into equally spaced bins."""
    return (
        pd.cut(x_var, bins=n_bins)
        .map(lambda x: (x.right + x.left) / 2)  # get interval midpoint
        .value_counts()
        .rename("Counts")
        .reset_index()
        .rename(columns={"index": x_var.name})
    )


def countplot_num(x_var: pd.Series) -> alt.Chart:
    """Plot counts by bins for numerical variables."""
    non_missing = ~x_var.isna()
    x_var = x_var[non_missing]

    x_var = trim_outliers(x_var)
    df_bins = get_bins(x_var)

    var_sanitized = vega_sanitize(
        x_var.name, datatype="Q"
    )  # format as quantitative type

    return alt.Chart(df_bins).mark_bar().encode(x=var_sanitized, y="Counts")


def countplot_cat(x_var: pd.Series, n_categories: int = 25) -> alt.Chart:
    """Plot counts by category for categorical variables."""
    var = x_var.name
    non_missing = ~x_var.isna()
    x_var = x_var[non_missing]

    df = (
        x_var.astype({var: "category"})
        .value_counts()
        .rename("Counts")
        .reset_index()
        .rename(columns={"index": var})
        .sort_values(by="Counts", ascending=False)
        .head(n_categories)
    )

    if is_categorical_dtype(x_var) == False:
        df[var] = df[var].cat.remove_unused_categories()
        ordered_categories = list(df[var].values)  # sort unordered uncategorical
    elif x_var.cat.ordered:
        ordered_categories = x_var.cat.categories
        df[var].cat.categories = ordered_categories
        df[var] = df[var].cat.remove_unused_categories()
        ordered_categories = list(df[var].cat.categories)
    else:
        df[var] = df[var].cat.remove_unused_categories()
        ordered_categories = list(df[var].values)  # sort unordered categories

    var_sanitized = vega_sanitize(var, datatype="N")  # format as nominal type

    return (
        alt.Chart(df)
        .mark_bar()
        .encode(x="Counts", y=alt.Y(var_sanitized, sort=ordered_categories))
    )


def countplot_datetime(x_var: pd.Series) -> alt.Chart:
    """Plot counts by bins for datetime variables."""
    non_missing = ~x_var.isna()
    x_var = x_var[non_missing]
    dateime_dtype = x_var.dtype

    x_var = trim_outliers(pd.to_numeric(x_var))
    df_bins = get_bins(x_var).astype({x_var.name: dateime_dtype})

    var_sanitized = vega_sanitize(x_var.name, datatype="T")  # format as datetime

    return alt.Chart(df_bins).mark_bar().encode(x=var_sanitized, y="Counts")



import operator
from collections import defaultdict
from typing import Any, Dict, Iterable, List

import numpy as np
import pandas
from pandas.api.types import (
    is_bool_dtype,
    is_categorical_dtype,
    is_float_dtype,
    is_integer_dtype,
    is_object_dtype,
)


def df_to_json_table(df, *args, **kwargs):
    """Create a json table from a pandas dataframe

    Parameters
    ----------
    df : pandas.Dataframe
        Input dataframe
    *args: passed on to df.to_dict
    **kwargs: passed on to df.to_dict

    Returns
    -------
    string
        json encoded and versioned table
    """

    version = "tktl v0.1"

    string = df.to_json(*args, orient="split", **kwargs, index=False)

    return string[:-1] + ',"version":"' + version + '"}'


def df_to_dict(df):
    """Create a dictionary from pandas dataframes
    Types are converted to native Python types using pandas' built-in
    mapping (see pandas.DataFrame.to_dict()).
    Parameters
    ----------
    df : pandas.Dataframe
        Input dataframe
    Returns
    -------
    dict
        Dictionary with variable names as keys and columns as lists.
    """
    dataset = {}
    df = df.replace({np.nan: None})
    for var, values in df.to_dict().items():
        dataset[var] = list(values.values())
    return dataset


def create_description(series, n_options=100) -> Dict:
    """Create an input description for a series for use in dropdowns

    Parameters
    ----------
    series : pandas.Series
        Pandas Series to be described
    n_options : int, optional
        Number of options for dropandown menus, by default 100

    Returns
    -------
    dict
        Description of the series
    """
    col_type = series.dtype

    if is_categorical_dtype(col_type):
        options = list(series.cat.categories)[:n_options]
        field_type = "category"

    elif is_object_dtype(col_type):
        value_counts = series.value_counts(dropna=True)
        options = value_counts.keys().to_list()[:n_options]
        field_type = "category"

    elif is_bool_dtype(col_type):
        options = [True, False]
        field_type = "bool"

    elif is_integer_dtype(col_type):
        options = None
        field_type = "integer"

    elif is_float_dtype(col_type):
        options = None
        field_type = "float"

    else:
        options = None
        field_type = str(col_type)

    input_description = {
        "name": series.name,
        "field_type": field_type,
        "options": options,
    }

    return input_description


def vega_sanitize(var, datatype=None):
    """Sanitize string for use as variable name in Vega-Lite/Altair

    For background, see: https://github.com/altair-viz/altair/issues/284
    """
    var = var.replace(".", "\\.")
    var = var.replace("[", "\\[")
    var = var.replace("]", "\\]")
    if datatype:
        var = var + ":" + datatype
    return var


def format_time(X, var):
    """Format datetime for vega plots"""

    timedelta = X[var].max() - X[var].min()
    if timedelta.days > 1:
        format = "%x"
    else:
        format = "%X"
    return format


def input_schema_to_pandas(value: Any, names: List[str]) -> pandas.DataFrame:
    if isinstance(value, pandas.DataFrame):
        value.columns = names
        return value
    else:
        return pandas.DataFrame(value, columns=names)


def batched_non_object_col_mode(
    x: Iterable[pandas.DataFrame],
    y: Iterable[pandas.Series],
    first: pandas.DataFrame,
    col: str,
):
    non_object_modes = defaultdict(int)
    values, counts = np.unique(first[col].dropna().values, return_counts=True)

    for v, c in zip(values, counts):
        non_object_modes[v] += c
    for batch, label in zip(x, y):
        batch = drop_nas_from_batch(batch, label)
        values, counts = np.unique(batch[col].dropna(), return_counts=True)
        for v, c in zip(values, counts):
            non_object_modes[v] += c

    items = non_object_modes.items()

    if items:
        return max(non_object_modes.items(), key=operator.itemgetter(1))[0]
    else:
        return None


def batched_numeric_col_mean(
    x: Iterable[pandas.DataFrame],
    y: Iterable[pandas.Series],
    first: pandas.DataFrame,
    col: str,
):
    running_sum, running_length = first[col].dropna().sum(), len(first[col].dropna())
    for batch, label in zip(x, y):
        batch = drop_nas_from_batch(batch, label)
        clean = batch[col].dropna()
        running_sum += clean.sum()
        running_length += len(clean)

    if running_length:
        return running_sum / running_length
    else:
        return float("NaN")


def drop_nas_from_batch(x: pandas.DataFrame, y: pandas.Series) -> pandas.DataFrame:
    not_missing = [i for i, v in enumerate(y) if not pandas.isna(v)]
    n_missing = len(y) - len(not_missing)
    if n_missing > 0:
        return x.iloc[not_missing]
    else:
        return x



import numpy as np
import pytest

from profiling.accuracy import binary_tpr_fpr


@pytest.mark.parametrize(
    "y, preds, tpr_expected, fpr_expected",
    [
        ([0, 0], [0, 0], [0, 0], [0, 1]),
        ([0, 0], [0, 1], [0, 0, 0], [0, 0.5, 1]),
        ([0, 1], [0, 0], [0, 1], [0, 1]),
        ([0, 1], [0, 1], [0, 1, 1], [0, 0, 1]),
        ([1, 1], [0, 1], [0, 0.5, 1], [0, 0, 0]),
    ],
)
def test_binary_tpr_fpr(y, preds, tpr_expected, fpr_expected):
    y = np.array(y)
    preds = np.array(preds)
    tpr, fpr = binary_tpr_fpr(y, preds)

    assert np.array_equal(tpr, tpr_expected)
    assert np.array_equal(fpr, fpr_expected)



from datetime import datetime

import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from profiling.sliceloss import slice_loss_cat, slice_loss_datetime, slice_loss_num


def predict(df):
    return pd.Series(data=np.ones(len(df)), index=df.index).astype(int)


X_unique_vals = pd.DataFrame(
    {
        "num": np.arange(1, 10),
        "cat": np.arange(1, 10),
        "time": pd.date_range("2021", freq="D", periods=9),
    }
).astype({"num": "int64", "cat": "category", "time": "datetime64[ns]"})

y_unique_vals = pd.Series(
    data=np.concatenate([np.zeros(4), np.ones(5)]), index=X_unique_vals.index
).astype(int)

pred_unique_vals = pd.DataFrame(
    {
        "num": X_unique_vals["num"].unique(),
        "cat": X_unique_vals["cat"].cat.categories,
        "time": X_unique_vals["time"].unique(),
        "Accuracy": np.concatenate([np.zeros(4), np.ones(5)]),
        "Counts": np.ones(9),
    },
).astype(
    {"num": "int64", "cat": "category", "time": "datetime64[ns]", "Counts": "int64"}
)


X_repeated_vals = pd.DataFrame(
    {
        "num": np.concatenate([[1] * 15, [2] * 30, [3] * 60, [4] * 30, [5] * 15]),
        "cat": np.concatenate([[1] * 15, [2] * 30, [3] * 60, [4] * 30, [5] * 15]),
        "time": np.concatenate(
            [
                [datetime(2021, 1, 1)] * 15,
                [datetime(2021, 1, 2)] * 30,
                [datetime(2021, 1, 3)] * 60,
                [datetime(2021, 1, 4)] * 30,
                [datetime(2021, 1, 5)] * 15,
            ]
        ),
    }
).astype({"num": "int64", "cat": "category", "time": "datetime64[ns]"})

y_repeated_vals = pd.Series(
    data=np.concatenate(
        [
            np.zeros(15),
            np.ones(15),
            np.zeros(15),
            np.ones(15),
            np.zeros(45),
            np.zeros(5),
            np.ones(25),
            np.ones(3),
            np.zeros(12),
        ]
    ),
    index=X_repeated_vals.index,
).astype(int)

pred_repeated_values = pd.DataFrame(
    {
        "num": X_repeated_vals["num"].unique(),
        "cat": X_repeated_vals["cat"].cat.categories,
        "time": X_repeated_vals["time"].unique(),
        "Accuracy": [0, 0.5, 0.25, 5 / 6, 0.2],
        "Counts": [15, 30, 60, 30, 15],
    }
).astype(
    {"num": "int64", "cat": "category", "time": "datetime64[ns]", "Counts": "int64"}
)


@pytest.mark.parametrize(
    "func, x, y, pred",
    [
        (predict, X_unique_vals, y_unique_vals, pred_unique_vals),
        (predict, X_repeated_vals, y_repeated_vals, pred_repeated_values),
    ],
)
def test_slice_loss_num(func, x, y, pred):

    data = slice_loss_num(func, x, y, "num", "Accuracy").data
    pred_num = pred[["num", "Accuracy", "Counts"]]
    assert_frame_equal(data, pred_num, check_dtype=False)


@pytest.mark.parametrize(
    "func, x, y, pred",
    [
        (predict, X_unique_vals, y_unique_vals, pred_unique_vals),
        (predict, X_repeated_vals, y_repeated_vals, pred_repeated_values),
    ],
)
def test_slice_loss_cat(func, x, y, pred):

    data = slice_loss_cat(func, x, y, "cat", "Accuracy").data.sort_values(by="cat")
    pred_cat = pred[["cat", "Accuracy", "Counts"]]
    assert_frame_equal(data, pred_cat)


@pytest.mark.parametrize(
    "func, x, y, pred",
    [
        (predict, X_unique_vals, y_unique_vals, pred_unique_vals),
        (predict, X_repeated_vals, y_repeated_vals, pred_repeated_values),
    ],
)
def test_slice_loss_datetime(func, x, y, pred):

    data = slice_loss_datetime(func, x, y, "time", "Accuracy").data.sort_values(
        by="time"
    )
    pred_time = pred[["time", "Accuracy", "Counts"]]
    assert_frame_equal(data, pred_time)
